#!/usr/bin/env python

from setuptools import setup, find_packages


setup(
    name="aiocached",
    version="0.3",
    description="Simple decorator to cache coroutine function results",
    long_description=open('README.md', encoding='utf8').read(),
    long_description_content_type='text/markdown',
    platforms="all",
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Development Status :: 5 - Production/Stable",
    ],
    license="Apache License, Version 2.0",
    author="Alexander Tikhonov",
    author_email="random.gauss@gmail.com",
    url='https://github.com/Fatal1ty/aiocached',
    packages=find_packages(exclude=("tests",)),
    python_requires=">=3.5",
)



from time import time


UNLIMITED = None


class Cache:
    def __init__(self, ttl=UNLIMITED):
        self.dictionary = {}
        self.ttl = ttl
        self.times = {}

    def __getitem__(self, item):
        if self.ttl is not UNLIMITED:
            self.remove_if_old(item)
        return self.dictionary[item]

    def __setitem__(self, key, value):
        self.dictionary[key] = value
        if self.ttl is not UNLIMITED:
            self.times[key] = time()

    def remove_if_old(self, item):
        t = self.times.get(item)
        if t and time() - t > self.ttl:
            del self.dictionary[item]
            del self.times[item]


class AsyncCache(Cache):
    def __init__(self, *args, **kwargs):
        super(AsyncCache, self).__init__(*args, **kwargs)
        self.futures = {}

    async def get(self, item):
        if self.ttl is not UNLIMITED:
            self.remove_if_old(item)
        try:
            return self.dictionary[item]
        except KeyError as e:
            fut = self.futures.get(item)
            if fut:
                await fut
                return fut.result()
            else:
                raise e from None

    def remove_if_old(self, item):
        t = self.times.get(item)
        if t and time() - t > self.ttl:
            del self.dictionary[item]
            del self.times[item]
            del self.futures[item]


__all__ = [
    'Cache',
    'AsyncCache',
    'UNLIMITED',
]



import asyncio
from functools import wraps

from .cache import AsyncCache, UNLIMITED


def cached(ttl=UNLIMITED):

    if callable(ttl):
        return cached(UNLIMITED)(ttl)

    def decorator(func):

        setattr(func, '__cache', AsyncCache(ttl))

        @wraps(func)
        def wrapped(*args, **kwargs):
            cache_key = (args, tuple(kwargs.items()))
            try:
                result = func.__cache[cache_key]
            except KeyError:
                try:
                    result = func(*args, **kwargs)
                except Exception as e:
                    raise e from None
                func.__cache[cache_key] = result
            return result

        @wraps(func)
        async def wrapped_coro(*args, **kwargs):
            cache_key = (args, tuple(kwargs.items()))
            try:
                result = await func.__cache.get(cache_key)
            except KeyError:
                try:
                    func.__cache.futures[cache_key] = asyncio.Future()
                    result = await func(*args, **kwargs)
                    func.__cache.futures[cache_key].set_result(result)
                except Exception as e:
                    fut = func.__cache.futures.pop(cache_key)
                    fut.cancel()
                    raise e from None
                func.__cache[cache_key] = result
            return result

        if asyncio.iscoroutinefunction(func):
            return wrapped_coro
        else:
            return wrapped

    return decorator


__all__ = [
    'cached',
]
